import logging

import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import TQDMProgressBar
from pytorch_lightning.utilities.warnings import PossibleUserWarning
import warnings

from .data import MUSEDataModule
from .modules import MUSE


def fit_predict(
        transcript_features: np.array,
        morphology_features: np.array,
        transcript_labels: np.array,
        morphology_labels: np.array,
        batch_size: int = 512,
        **kwargs,
):

    warnings.filterwarnings("ignore", category=PossibleUserWarning)
    warnings.filterwarnings("ignore", ".*value needs to be floating point.*")
    muse = MUSE(transcript_features.shape[1],
                morphology_features.shape[1], **kwargs)

    muse_datamodule = MUSEDataModule(
        transcript_features, morphology_features, transcript_labels, morphology_labels, batch_size)

    trainer = pl.Trainer(accelerator='gpu', devices=1, max_epochs=-1,
                         callbacks=[TQDMProgressBar()], log_every_n_steps=1)

    trainer.fit(muse, muse_datamodule,)

    out = trainer.predict(muse, muse_datamodule,)

    # Combine outputs
    z = []
    x_hat = []
    y_hat = []
    latent_x = []
    latent_y = []

    for batch in out:
        z.append(batch['z'])
        x_hat.append(batch['x_hat'])
        y_hat.append(batch['y_hat'])
        latent_x.append(batch['latent_x'])
        latent_y.append(batch['latent_y'])

    z = torch.cat(z, dim=0)
    x_hat = torch.cat(x_hat, dim=0)
    y_hat = torch.cat(y_hat, dim=0)
    latent_x = torch.cat(latent_x, dim=0)
    latent_y = torch.cat(latent_y, dim=0)

    return z, x_hat, y_hat, latent_x, latent_y



from typing import Optional

import numpy as np
import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader, Dataset


class MUSEFeaturesDataset(Dataset):
    def __init__(self,
                 transcript_features: np.array,
                 morphology_features: np.array) -> None:

        self.morphology_features = morphology_features
        self.transcript_features = transcript_features

        assert morphology_features.shape[0] == transcript_features.shape[
            0], "The first dimension of both features should be equal to the sample size"

    def get_sample_size(self):
        return self.morphology_features.shape[0]

    def __getitem__(self, index: int):
        return (
            torch.tensor(self.transcript_features[index]).float(),
            torch.tensor(self.morphology_features[index]).float(),)

    def __len__(self):
        return self.transcript_features.shape[0]


class MUSELabelsDataset(Dataset):
    def __init__(self,
                 transcript_labels: np.array,
                 morphology_labels: np.array) -> None:

        self.transcript_labels = list(transcript_labels)
        self.morphology_labels = list(morphology_labels)

        assert len(self.transcript_labels) == len(self.morphology_labels)

    def __getitem__(self, index: int):
        return (
            torch.tensor(self.transcript_labels[index]).float(),
            torch.tensor(self.morphology_labels[index]).float(),)

    def __len__(self):
        return len(self.transcript_labels)


class MUSEDataModule(pl.LightningDataModule):
    def __init__(self,
                 transcript_features: np.array,
                 morphology_features: np.array,
                 transcript_labels: np.array,
                 morphology_labels: np.array,
                 batch_size: int = 512):

        super().__init__()
        self.save_hyperparameters()
        self.batch_size = batch_size

    def update_labels(self, transcript_labels, morphology_labels):
        """Re define labels dataset with the newly computed labels

        Args:
            x_labels (_type_): _description_
            y_labels (_type_): _description_
        """

        # Create new iterator for the updated labels
        self.labels_dataset = MUSELabelsDataset(
            transcript_labels, morphology_labels)

    def setup(self, stage: Optional[str] = None):
        self.features_dataset = MUSEFeaturesDataset(
            self.hparams.transcript_features, self.hparams.morphology_features)
        self.labels_dataset = MUSELabelsDataset(
            self.hparams.transcript_labels, self.hparams.morphology_labels)

    def train_dataloader(self):
        # only load features dataloader when
        features_loader = DataLoader(
            self.features_dataset, batch_size=self.batch_size, num_workers=6, drop_last=False, shuffle=False)

        labels_loader = DataLoader(
            self.labels_dataset, batch_size=self.batch_size, num_workers=6, drop_last=False, shuffle=False)

        loaders = {"features": features_loader, "labels": labels_loader}
        return loaders

    def predict_dataloader(self):
        # only load features dataloader when
        features_loader = DataLoader(
            self.features_dataset, batch_size=self.batch_size, num_workers=6)

        return features_loader



import torch
from torch import nn
from torch.nn.init import kaiming_normal_, zeros_


def init_weights(m):

    for l in m.children():
        if isinstance(l, nn.Linear):
            kaiming_normal_(l.weight)

            if l.bias is not None:
                zeros_(l.bias)


class MultiViewEncoder(nn.Module):

    def __init__(self, in_features_x: int, in_features_y: int, sm_latent_dim: int = 128, joint_latent_dim: int = 100) -> None:
        super().__init__()

        self.in_features_x = in_features_x
        self.in_features_y = in_features_y
        self.sm_latent_dim = sm_latent_dim
        self.joint_latent_dim = joint_latent_dim

        self.x_encoder = Encoder(
            in_features=self.in_features_x, hidden_dim=self.sm_latent_dim)
        self.y_encoder = Encoder(
            in_features=self.in_features_y, hidden_dim=self.sm_latent_dim)
        self.joint_encoder = Encoder(
            in_features=self.sm_latent_dim*2, hidden_dim=self.joint_latent_dim)

        self.apply(init_weights)

    def forward(self, x, y):
        x_latent = self.x_encoder(x)
        y_latent = self.y_encoder(y)

        joint_latent = torch.concat((x_latent, y_latent), dim=1)
        joint_latent = self.joint_encoder(joint_latent)

        return joint_latent, x_latent, y_latent


class Encoder(nn.Module):
    def __init__(self, in_features: int, hidden_dim: int = 128) -> None:
        super().__init__()

        self.in_features = in_features
        self.hidden_dim = hidden_dim

        self.fc1 = nn.Linear(in_features=self.in_features,
                             out_features=self.hidden_dim)
        self.fc2 = nn.Linear(in_features=self.hidden_dim,
                             out_features=self.hidden_dim)

        self.apply(init_weights)

    def forward(self, x):
        x = self.fc1(x)
        x = self.fc2(x)

        return x


class Decoder(nn.Module):
    def __init__(self, in_features: int, out_features: int, hidden_dim: int = 128) -> None:
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.hidden_dim = hidden_dim

        self.fc1 = nn.Linear(in_features=self.in_features,
                             out_features=hidden_dim)
        self.fc2 = nn.Linear(in_features=self.hidden_dim,
                             out_features=hidden_dim)
        self.fc3 = nn.Linear(in_features=self.hidden_dim,
                             out_features=out_features)

        self.apply(init_weights)

    def forward(self, x):
        x = self.fc1(x)
        x = self.fc2(x)
        x = self.fc3(x)

        return x


class WeightMatrix(torch.nn.Module):
    def __init__(self, joint_latent_dim: int = 100):
        super().__init__()
        self.joint_latent_dim = joint_latent_dim

        self.weight = torch.nn.Parameter(
            torch.zeros((joint_latent_dim, joint_latent_dim)))
        self.bias = None

        kaiming_normal_(self.weight)

    def fnorm(self):
        return torch.norm(self.weight, p='fro')

    def forward(self, x):
        return torch.matmul(x, self.weight)


class MUSELoss:
    def __init__(self, weight_penalty: float, triplet_lambda: float):
        self.weight_penality = weight_penalty
        self.triplet_lambda = triplet_lambda

    def __call__():
        return



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="muse-pytorch",
    version="0.0.3",
    author="Nicola Occelli",
    author_email="nicola.occelli@ulb.be",
    description="Pytorch version of MUSE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Nico995/MUSE-pytorch",
    project_urls={
        "Bug Tracker": "https://github.com/Nico995/MUSE-pytorch/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: GPU :: NVIDIA CUDA :: 10.2",
        "Natural Language :: English",
    ],
    package_dir={"": "muse_pytorch"},
    packages=setuptools.find_packages(where="muse_pytorch"),
    python_requires=">=3.8.5",
)



import torch
from online_triplet_loss.losses import batch_hard_triplet_loss


class ReconstructionLoss:
    def __init__(self, ):
        pass

    def __call__(self, x, x_hat):
        return torch.mean(torch.norm(x_hat - x, p=2))


class ZIReconstructionLoss:
    def __init__(self, ):
        pass

    def __call__(self, x, x_hat):
        x_mask = torch.sign(x)
        return torch.sum(torch.norm(torch.multiply(x_mask, (x_hat - x)), p=2)) / torch.sum(x_mask)


class TripletLoss:
    def __init__(self, margin):
        self.margin = margin

    def __call__(self, labels, embeddings):
        return batch_hard_triplet_loss(labels, embeddings, margin=self.margin)



import os
from typing import Tuple

import numpy as np
import phenograph
import pytorch_lightning as pl
import torch
from scipy.spatial.distance import pdist

from .blocks import Decoder, MultiViewEncoder, WeightMatrix
from .losses import ReconstructionLoss, TripletLoss, ZIReconstructionLoss
from .utils import block_print, enable_print


class MUSE(pl.LightningModule):
    def __init__(self,
                 ts_input_dim: int = 500,
                 ml_input_dim: int = 2048,
                 latent_dim: int = 128,
                 joint_latent_dim: int = 100,
                 lambda_reg: float = 5.,
                 lambda_sup: float = 5.,
                 lr: float = 1e-04,
                 init_epochs: int = 100,
                 refine_epochs: int = 100,
                 cluster_epochs: int = 100,
                 cluster_update_epoch: int = 50
                 ) -> None:
        """MUSE implementation in pytorch-lightning.

        This implementation tries to stay as close as possible to the original one, but it is not always possible.
        For convenience, the training has been divided in 3 phases that I arbitrarily called initialziation, refinement and clustering.

         - The initialziation phase only aims at optimizing the reconstruction part of the network. 
        For this reason in the first phase no self-supervised loss or clustering is computed.
        At the end of the initialization phase the margin of the triplet loss is estimated and the self-supervised loss is activated.
         - The refinement phase trains both on the recosntruction and self-supervised loss. 
         Although, at this time no further clustering of the samples is computed, and the loss uses only the reference labels
          - The clustering phase also refines the initial clusters, by running a clustering algorithm every cluster_update_epoch epochs.

        Args:
            ts_input_dim (int, optional): Number of features in the transcript modality. Defaults to 500.
            ml_input_dim (int, optional): Number of features in the morphology modality. Defaults to 2048.
            latent_dim (int, optional): Size of the latent dimension of the encoded single modalities (hx and yx in the paper). Defaults to 128.
            joint_latent_dim (int, optional): Size of the latent dimension of the joint modalities (z in the paper). Defaults to 100.
            lambda_reg (float, optional): Wegith for the regularization term in the loss function (lambda_regularization in the paper). Defaults to 5..
            lambda_sup (float, optional): Wegith for the self-supervised term in the loss function (lambda_supervise in the paper). Defaults to 5..
            lr (float, optional): Learning rate for the model. Defaults to 1e-04.
            init_epochs (int, optional): Number of epochs of the initialiation part of training. Defaults to 100.
            refine_epochs (int, optional): Number of epochs of the refining part of training. Defaults to 100.
            cluster_epochs (int, optional): Number of epochs for the clustering part of training. Defaults to 100.
            cluster_update_epoch (int, optional): Number of epoch after which the clustering on the single modalities are updated. Defaults to 50.
        """
        super().__init__()

        self.save_hyperparameters()
        self.mve = MultiViewEncoder(
            in_features_x=ts_input_dim,
            in_features_y=ml_input_dim,
            sm_latent_dim=latent_dim,
            joint_latent_dim=joint_latent_dim
        )

        self.wm_x = WeightMatrix(joint_latent_dim=joint_latent_dim)
        self.wm_y = WeightMatrix(joint_latent_dim=joint_latent_dim)

        self.decoder_x = Decoder(
            in_features=joint_latent_dim,
            out_features=ts_input_dim
        )

        self.decoder_y = Decoder(
            in_features=joint_latent_dim,
            out_features=ml_input_dim
        )

        # Triplet loss function
        self.triplet_loss = TripletLoss(margin=0)
        # Reconstruction loss function
        self.rec_loss = ReconstructionLoss()
        # Zero Inflated reconstruction loss function
        self.zi_rec_loss = ZIReconstructionLoss()

        self.phase = 'init'

    # Move this function into the datamodule

    def _cluster_latent(self, latent_x: torch.Tensor, latent_y: torch.Tensor) -> Tuple[np.array]:
        """Performs clustering over latent representations of single modalities.

        Args:
            latent_x (torch.Tensor): Latent representation of the transcript modality
            latent_y (torch.Tensor): Latent representation of the morphological modality

        Returns:
            Tuple[np.array]: containing the cluster labels for each sample in both modalities.
        """

        block_print()
        labels_x, _, Q_x = phenograph.cluster(latent_x.detach().cpu())
        labels_y, _, Q_y = phenograph.cluster(latent_y.detach().cpu())
        enable_print()

        print(f'Clustering done! Q scores for x and y are {Q_x}\t{Q_y} ')

        return labels_x, labels_y

    def estimate_margin(self, z: torch.Tensor) -> None:
        """Compute margin estimation and redefine triplet loss with the newly computer margin

        Args:
            z (torch.Tensor): Joint latent representation 
        """

        latent_pd_matrix = pdist(z.detach().cpu(), 'euclidean')
        latent_pd_sort = np.sort(latent_pd_matrix)
        select_top_n = np.int(latent_pd_sort.size * 0.2)
        margin_estimate = np.median(
            latent_pd_sort[-select_top_n:]) - np.median(latent_pd_sort[:select_top_n])

        # Store margin and re-initialize loss
        self.hparams.triplet_margin = margin_estimate
        self.triplet_loss = TripletLoss(margin=margin_estimate)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(), lr=self.hparams.lr)
        return optimizer

    def predict_step(self, batch, batch_idx):

        x, y = batch

        z, latent_x, latent_y = self.mve(x, y)

        z_x, z_y = self.wm_x(z), self.wm_y(z)

        x_hat,  y_hat = self.decoder_x(z_x), self.decoder_y(z_y)

        return {'z': z, 'x_hat': x_hat, 'y_hat': y_hat, 'latent_x': latent_x, 'latent_y': latent_y}

    def training_step(self, batch, batch_idx):

        # I will try to keep the same naming as the original implementation
        (x, y), (labels_x, labels_y) = batch['features'], batch['labels']

        # Get Single Modality and joint latent spaces from the multi-view encoder
        z, latent_x, latent_y = self.mve(x, y)

        # Get modality specific latent space from joint latent space
        z_x, z_y = self.wm_x(z), self.wm_y(z)

        # Decode the modality specific latent spaces to reconstruct original features
        x_hat,  y_hat = self.decoder_x(z_x), self.decoder_y(z_y)

        #==== Loss ====#
        # Frobenious norm minimization
        sparse_penality = self.wm_x.fnorm() + self.wm_y.fnorm()

        # Reconstruction loss for each single modality
        rec_loss = self.rec_loss(x, x_hat) + self.rec_loss(y, y_hat)

        # Triplet loss (not computed during init phase)
        if self.phase != 'init':
            triplet_loss = self.triplet_loss(labels_x, latent_x) \
                + self.triplet_loss(labels_y, latent_y)
        else:
            triplet_loss = 0

        loss = rec_loss + self.hparams.lambda_reg * \
            sparse_penality + self.hparams.lambda_sup * triplet_loss

        self.log(f"loss",
                 {"reconstruction": rec_loss, "triplet": triplet_loss, "total": loss},
                 prog_bar=False, on_step=False, on_epoch=True)

        return {'loss': loss, 'latent_x': latent_x, 'latent_y': latent_y, 'z': z}

    def training_epoch_end(self, outputs) -> None:

        # Switch training phase according to input parameters
        # End of init phase -> refine phase
        if self.trainer.current_epoch + 1 == self.hparams.init_epochs:
            self.phase = 'refine'

            # At the end of the refine phase we can estimate the margin for the triplet loss
            z = torch.cat([batch['z'] for batch in outputs])
            self.estimate_margin(z)

            print('Initalization phase completed. Moving to Refining phase.')

        elif self.trainer.current_epoch + 1 == self.hparams.init_epochs + self.hparams.refine_epochs:
            self.phase = 'cluster'

            print('Refining phase completed. Moving to Clustering phase.')

        # When running cluster, we must update our labels
        if self.phase == 'cluster':

            cluster_epoch = self.trainer.current_epoch - \
                (self.hparams.init_epochs + self.hparams.refine_epochs)

            # Update clusters every cluster_update_epoch epochs for the start of the cluster phase
            if cluster_epoch % self.hparams.cluster_update_epoch == 0:
                latent_x = torch.cat([batch['latent_x']
                                      for batch in outputs])
                latent_y = torch.cat([batch['latent_y']
                                      for batch in outputs])

                x_labels, y_labels = self._cluster_latent(
                    latent_x, latent_y)

                self.trainer.datamodule.update_labels(x_labels, y_labels)
                self.trainer.reset_train_dataloader()

        if self.trainer.current_epoch + 1 == self.hparams.init_epochs + self.hparams.refine_epochs + self.hparams.cluster_epochs:
            self.trainer.should_stop = True
            self.z = torch.cat([batch['z'] for batch in outputs])
            torch.save(self.z, os.path.join(self.trainer.log_dir, "z.pt"))
            print('Clustering phase completed. Training done.')
